@@ -3,8 +3,10 @@ use super::DecoderResult;
 
 pub mod dcl;
 pub mod imaadpcm;
+pub mod lzs;
 pub mod lzss;
 pub mod path;
+pub mod rnc;
 
 pub trait ReadVGAPal {
     fn read_vga_pal(&mut self, pal: &mut [u8; 768]) -> DecoderResult<()>;
