@@ -402,6 +402,13 @@ impl<S: Copy> CodebookDescReader<S> for FullCodebookDescReader<S> {
     fn len(&mut self) -> usize { self.data.len() }
 }
 
+impl<S: Copy> CodebookDescReader<S> for Vec<FullCodebookDesc<S>> {
+    fn bits(&mut self, idx: usize) -> u8  { self[idx].bits }
+    fn code(&mut self, idx: usize) -> u32 { self[idx].code }
+    fn sym (&mut self, idx: usize) -> S   { self[idx].sym  }
+    fn len(&mut self) -> usize { self.as_slice().len() }
+}
+
 /// Codebook description that stores a list of codewords and their value is equal to the index.
 pub struct ShortCodebookDescReader {
     data: Vec<ShortCodebookDesc>,
