@@ -4,6 +4,9 @@ use std::io::BufWriter;
 use crate::io::byteio::*;
 use super::*;
 
+const MAX_RIFF_BLOCKS: usize = 4; // ~4GB file should be enough for game videos
+const MAX_RIFF_SIZE: u64 = 900 << 20; // 900MB of payload plus index data will hopefully fit in 1GB chunk limit
+
 #[cfg(debug_assertions)]
 macro_rules! validate {
     ($a:expr) => { if !$a { println!("check failed at {}:{}", file!(), line!()); return Err(EncoderError::InvalidData); } };
@@ -16,7 +19,7 @@ macro_rules! validate {
 struct IndexEntry {
     tag:    [u8; 4],
     size:   u32,
-    offs:   u32,
+    offs:   u64,
 }
 
 struct StreamStats {
@@ -25,9 +28,11 @@ struct StreamStats {
     has_pal:    bool,
     has_pc:     bool,
     nframes:    u32,
+    nticks:     u32,
     max_size:   u32,
     strh_pos:   u64,
     is_video:   bool,
+    indx_pos:   u64,
 }
 
 struct AVIWriter {
@@ -36,6 +41,99 @@ struct AVIWriter {
     streams:    Vec<StreamInfo>,
     index:      Vec<IndexEntry>,
     data_pos:   u64,
+    odml_pos:   u64,
+    riff_blk:   usize,
+    blk_start:  u64,
+    wrote_hdr:  bool,
+}
+
+impl AVIWriter {
+    fn patch_stream_stats(&mut self) -> EncoderResult<()> {
+        let mut bw = ByteWriter::new(&mut self.fr);
+        let mut max_frames = 0;
+        let mut max_size = 0;
+        for sstat in self.sstats.iter() {
+            max_frames = max_frames.max(sstat.nframes);
+            max_size = max_size.max(sstat.max_size);
+            if sstat.has_pc {
+                bw.seek(SeekFrom::Start(sstat.strh_pos))?;
+                bw.write_u32le(0x00010000)?;
+            }
+            bw.seek(SeekFrom::Start(sstat.strh_pos + 0x18))?;
+            bw.write_u32le(if sstat.is_video { sstat.nframes } else { 0 })?;
+            bw.write_u32le(sstat.max_size)?;
+        }
+        bw.seek(SeekFrom::Start(0x30))?;
+        bw.write_u32le(max_frames)?;
+        bw.seek(SeekFrom::Current(8))?;
+        bw.write_u32le(max_size)?;
+        Ok(())
+    }
+    fn flush_riff_block(&mut self) -> EncoderResult<()> {
+        let mut bw = ByteWriter::new(&mut self.fr);
+
+        for (sidx, (stream, sstats)) in self.streams.iter().zip(self.sstats.iter_mut()).enumerate() {
+            let idx_pos = bw.tell();
+
+            let tag0 = b'0' + ((sidx / 10) as u8);
+            let tag1 = b'0' + ((sidx % 10) as u8);
+
+            let nentries = self.index.iter().filter(|idx| idx.tag[0] == tag0 && idx.tag[1] == tag1).count();
+
+            let tag = [b'i', b'x', tag0, tag1];
+            bw.write_buf(&tag)?;
+            bw.write_u32le((6 * 4 + nentries * 8) as u32)?;
+            bw.write_u16le(2)?;
+            bw.write_byte(0)?;
+            bw.write_byte(1)?; // AVI_INDEX_OF_CHUNKS
+            let mut chunk_id = [tag0, tag1, 0, 0];
+            match stream {
+                StreamInfo::Video(_) => {
+                    chunk_id[2] = b'd';
+                    chunk_id[3] = b'b';
+                },
+                StreamInfo::Audio(_) => {
+                    chunk_id[2] = b'w';
+                    chunk_id[3] = b'b';
+                },
+                _ => unreachable!(),
+            }
+            bw.write_u32le(nentries as u32)?;
+            bw.write_buf(&chunk_id)?;
+            bw.write_u64le(self.blk_start)?;
+            bw.write_u32le(0)?;
+
+            for entry in self.index.iter().filter(|idx| idx.tag[0] == tag0 && idx.tag[1] == tag1) {
+                bw.write_u32le((entry.offs - self.blk_start + 8) as u32)?;
+                bw.write_u32le(entry.size)?;
+            }
+
+            let idx_end = bw.tell();
+
+            bw.seek(SeekFrom::Start(sstats.indx_pos + 8 * 4 + (self.riff_blk as u64) * 16))?;
+            bw.write_u64le(idx_pos)?;
+            bw.write_u32le((idx_end - idx_pos) as u32)?;
+            bw.write_u32le(sstats.nticks)?;
+            sstats.nticks = 0;
+
+            bw.seek(SeekFrom::Start(idx_end))?;
+        }
+
+        patch_size(&mut bw, self.blk_start + 8)?;
+        patch_size(&mut bw, self.data_pos)?;
+
+        self.index.clear();
+
+        self.riff_blk += 1;
+        self.blk_start = bw.tell();
+        self.wrote_hdr = false;
+
+        if self.riff_blk == 1 {
+            self.patch_stream_stats()?;
+        }
+
+        Ok(())
+    }
 }
 
 impl OutputWriter for AVIWriter {
@@ -123,12 +221,14 @@ impl OutputWriter for AVIWriter {
                 has_pc:     false,
                 pal_pos:    0,
                 nframes:    0,
+                nticks:     0,
                 max_size:   0,
                 strh_pos:   0,
                 is_video:   false,
+                indx_pos:   0,
             });
         }
-        for (stream, sstat) in self.streams.iter().zip(self.sstats.iter_mut()) {
+        for (sidx, (stream, sstat)) in self.streams.iter().zip(self.sstats.iter_mut()).enumerate() {
             let strl_pos = bw.tell() + 8;
             bw.write_buf(b"LIST\0\0\0\0strlstrh")?;
             bw.write_u32le(56)?; // strh size
@@ -231,8 +331,68 @@ impl OutputWriter for AVIWriter {
                 _ => unreachable!(),
             };
             patch_size(&mut bw, strf_pos)?;
+            /* OpenDML */ {
+                sstat.indx_pos = bw.tell();
+                bw.write_buf(b"indx")?;
+                bw.write_u32le((6 * 4 + MAX_RIFF_BLOCKS * 16) as u32)?;
+                bw.write_u16le(4)?;
+                bw.write_byte(0)?;
+                bw.write_byte(0)?; // AVI_INDEX_OF_INDEXES
+                bw.write_u32le(0)?; // num entries
+                let mut ix_tag = [b'0' + ((sidx / 10) as u8), b'0' + ((sidx % 10) as u8), 0, 0];
+                match stream {
+                    StreamInfo::Video(ref _vinfo) => {
+                        ix_tag[2] = b'd';
+                        ix_tag[3] = b'b';
+                    },
+                    StreamInfo::Audio(ref _ainfo) => {
+                        ix_tag[2] = b'w';
+                        ix_tag[3] = b'b';
+                    },
+                    _ => unreachable!(),
+                }
+                bw.write_buf(&ix_tag)?;
+                for _ in 0..3 {
+                    bw.write_u32le(0)?;
+                }
+                for _ in 0..MAX_RIFF_BLOCKS {
+                    bw.write_u64le(0)?;
+                    bw.write_u32le(0)?;
+                    bw.write_u32le(0)?;
+                }
+
+                /*if let StreamInfo::Video(ref vinfo) = stream {
+                    bw.write_buf(b"vprp")?;
+                    bw.write_u32le(0x44)?;
+                    bw.write_u32le(0)?; // video format token
+                    bw.write_u32le(0)?; // video standard
+                    bw.write_u32le((vinfo.tb_den + vinfo.tb_num / 2) / vinfo.tb_num)?; // refresh rate
+                    bw.write_u32le(vinfo.width as u32)?;
+                    bw.write_u32le(vinfo.height as u32)?;
+                    bw.write_u32le(0x00010001)?; // aspect ratio
+                    bw.write_u32le(vinfo.width as u32)?;
+                    bw.write_u32le(vinfo.height as u32)?;
+                    bw.write_u32le(1)?; // single field
+                    bw.write_u32le(vinfo.height as u32)?;
+                    bw.write_u32le(vinfo.width as u32)?;
+                    bw.write_u32le(vinfo.height as u32)?;
+                    bw.write_u32le(vinfo.width as u32)?;
+                    bw.write_u32le(0)?;
+                    bw.write_u32le(0)?;
+                    bw.write_u32le(0)?;
+                    bw.write_u32le(0)?;
+                }*/
+            }
             patch_size(&mut bw, strl_pos)?;
         }
+        /* OpenDML */ {
+            self.odml_pos = bw.tell();
+            bw.write_buf(b"LIST")?;
+            bw.write_u32le(16)?;
+            bw.write_buf(b"odmldmlh")?;
+            bw.write_u32le(4)?;
+            bw.write_u32le(0)?;
+        }
         patch_size(&mut bw, hdrl_pos)?;
 
         self.data_pos = bw.tell() + 8;
@@ -245,16 +405,38 @@ impl OutputWriter for AVIWriter {
 
         let mut bw = ByteWriter::new(&mut self.fr);
 
-        if bw.tell() > (1 << 31) || self.sstats[stream_no].nframes > (1 << 20) {
+        if self.sstats[stream_no].nframes > (1 << 20) {
             return Err(EncoderError::ContainerFull);
         }
+        if bw.tell() > self.blk_start + MAX_RIFF_SIZE {
+            self.flush_riff_block()?;
+            bw = ByteWriter::new(&mut self.fr);
+        }
+
+        if !self.wrote_hdr {
+            if self.riff_blk >= MAX_RIFF_BLOCKS {
+                return Err(EncoderError::ContainerFull);
+            }
+            bw.seek(SeekFrom::Start(self.blk_start))?;
+
+            bw.write_buf(b"RIFF")?;
+            bw.write_u32le(0)?;
+            bw.write_buf(b"AVIX")?;
+            bw.write_buf(b"LIST")?;
+            bw.write_u32le(0)?;
+            self.data_pos = bw.tell();
+            bw.write_buf(b"movi")?;
+
+            self.wrote_hdr = true;
+        }
 
         let mut tag = [b'0' + ((stream_no / 10) as u8), b'0' + ((stream_no % 10) as u8), b' ', b' '];
-        let offs = bw.tell() as u32;
+        let offs = bw.tell();
 
         self.sstats[stream_no].nframes += 1;
         match (&self.streams[stream_no], &frame) {
             (StreamInfo::Video(ref vinfo), Frame::VideoPal(ref data, ref pal)) => {
+                self.sstats[stream_no].nticks += 1;
                 validate!(vinfo.bpp == 8);
                 if !self.sstats[stream_no].has_pal {
                     bw.seek(SeekFrom::Start(u64::from(self.sstats[stream_no].pal_pos)))?;
@@ -264,7 +446,7 @@ impl OutputWriter for AVIWriter {
                         bw.write_byte(clr[0])?;
                         bw.write_byte(0)?;
                     }
-                    bw.seek(SeekFrom::Start(u64::from(offs)))?;
+                    bw.seek(SeekFrom::Start(offs))?;
                     self.sstats[stream_no].pal.copy_from_slice(pal);
                     self.sstats[stream_no].has_pal = true;
                 } else if &self.sstats[stream_no].pal != pal {
@@ -282,13 +464,13 @@ impl OutputWriter for AVIWriter {
 
                 tag[2] = b'd';
                 tag[3] = b'b';
-                let offs = bw.tell() as u32;
-                let size = data.len() as u32;
+                let stride = vinfo.width;
+                let pad = vec![0; if (stride & 3) != 0 { 4 - (stride & 3) } else { 0 }];
+                let offs = bw.tell();
+                let size = ((stride + pad.len()) * vinfo.height) as u32;
                 self.index.push(IndexEntry { tag, offs, size });
                 bw.write_buf(&tag)?;
                 bw.write_u32le(size)?;
-                let stride = vinfo.width;
-                let pad = vec![0; if (stride & 3) != 0 { 4 - (stride & 3) } else { 0 }];
                 for line in data.chunks_exact(vinfo.width).rev() {
                     bw.write_buf(line)?;
                     bw.write_buf(&pad)?;
@@ -296,12 +478,13 @@ impl OutputWriter for AVIWriter {
                 self.sstats[stream_no].max_size = self.sstats[stream_no].max_size.max(size);
             },
             (StreamInfo::Video(ref vinfo), Frame::VideoRGB16(ref data)) => {
+                self.sstats[stream_no].nticks += 1;
                 validate!(vinfo.bpp == 15 || vinfo.bpp == 16);
                 match vinfo.bpp {
                     15 => {
                         tag[2] = b'd';
                         tag[3] = b'b';
-                        let size = (data.len() * 2) as u32;
+                        let size = (((vinfo.width + 1) & !1) * vinfo.height * 2) as u32;
                         self.index.push(IndexEntry { tag, offs, size });
                         bw.write_buf(&tag)?;
                         bw.write_u32le(size)?;
@@ -317,17 +500,17 @@ impl OutputWriter for AVIWriter {
                     16 => {
                         tag[2] = b'd';
                         tag[3] = b'b';
-                        let size = (data.len() * 3) as u32;
+                        let stride = vinfo.width * 3;
+                        let pad = vec![0; if (stride & 3) != 0 { 4 - (stride & 3) } else { 0 }];
+                        let size = ((stride + pad.len()) * vinfo.height) as u32;
                         self.index.push(IndexEntry { tag, offs, size });
                         bw.write_buf(&tag)?;
                         bw.write_u32le(size)?;
-                        let stride = vinfo.width * 3;
-                        let pad = vec![0; if (stride & 3) != 0 { 4 - (stride & 3) } else { 0 }];
                         for line in data.chunks_exact(vinfo.width).rev() {
                             for &pix in line.iter() {
-                                let r = ((pix >> 11) & 0x1F) as u8;
+                                let b = ((pix >> 11) & 0x1F) as u8;
                                 let g = ((pix >>  5) & 0x3F) as u8;
-                                let b = ( pix        & 0x1F) as u8;
+                                let r = ( pix        & 0x1F) as u8;
                                 bw.write_byte((r << 3) | (r >> 2))?;
                                 bw.write_byte((g << 2) | (g >> 4))?;
                                 bw.write_byte((b << 3) | (b >> 2))?;
@@ -340,21 +523,27 @@ impl OutputWriter for AVIWriter {
                 }
             },
             (StreamInfo::Video(ref vinfo), Frame::VideoRGB24(ref data)) => {
+                self.sstats[stream_no].nticks += 1;
                 validate!(vinfo.bpp == 24);
                 tag[2] = b'd';
                 tag[3] = b'b';
-                let size = data.len() as u32;
+                let stride = vinfo.width * 3;
+                let pad = vec![0; if (stride & 3) != 0 { 4 - (stride & 3) } else { 0 }];
+                let size = ((stride + pad.len()) * vinfo.height) as u32;
                 self.index.push(IndexEntry { tag, offs, size });
                 bw.write_buf(&tag)?;
                 bw.write_u32le(size)?;
-                let stride = vinfo.width * 3;
-                let pad = vec![0; if (stride & 3) != 0 { 4 - (stride & 3) } else { 0 }];
                 for line in data.chunks_exact(vinfo.width * 3).rev() {
-                    bw.write_buf(line)?;
+                    for pix in line.chunks_exact(3) {
+                        bw.write_byte(pix[2])?;
+                        bw.write_byte(pix[1])?;
+                        bw.write_byte(pix[0])?;
+                    }
                     bw.write_buf(&pad)?;
                 }
             },
             (StreamInfo::Audio(ref ainfo), Frame::AudioU8(ref data)) => {
+                self.sstats[stream_no].nticks += (data.len() / usize::from(ainfo.channels)) as u32;
                 validate!(ainfo.sample_type == AudioSample::U8);
                 tag[2] = b'w';
                 tag[3] = b'b';
@@ -366,6 +555,7 @@ impl OutputWriter for AVIWriter {
                 self.sstats[stream_no].max_size = self.sstats[stream_no].max_size.max(size);
             },
             (StreamInfo::Audio(ref ainfo), Frame::AudioS16(ref data)) => {
+                self.sstats[stream_no].nticks += (data.len() / usize::from(ainfo.channels) / 2) as u32;
                 validate!(ainfo.sample_type == AudioSample::S16);
                 tag[2] = b'w';
                 tag[3] = b'b';
@@ -387,7 +577,25 @@ impl OutputWriter for AVIWriter {
         Ok(())
     }
     fn finish(&mut self) -> EncoderResult<()> {
-        if !self.index.is_empty() {
+        if self.riff_blk == 0 && self.index.is_empty() {
+            return Err(EncoderError::EmptyOutput);
+        }
+        if self.riff_blk > 0 {
+            self.flush_riff_block()?;
+            let mut bw = ByteWriter::new(&mut self.fr);
+            let mut max_frames = 0;
+            for sstat in self.sstats.iter() {
+                max_frames = max_frames.max(sstat.nframes);
+
+                bw.seek(SeekFrom::Start(sstat.indx_pos + 12))?;
+                bw.write_u32le(self.riff_blk as u32)?;
+            }
+            bw.seek(SeekFrom::Start(self.odml_pos + 20))?;
+            bw.write_u32le(max_frames)?;
+
+            bw.flush()?;
+            Ok(())
+        } else if !self.index.is_empty() {
             let mut bw = ByteWriter::new(&mut self.fr);
 
             patch_size(&mut bw, self.data_pos)?;
@@ -397,29 +605,23 @@ impl OutputWriter for AVIWriter {
                 for item in self.index.iter() {
                     bw.write_buf(&item.tag)?;
                     bw.write_u32le(if &item.tag[2..] == b"pc" { 0x100 } else { 0x10 })?;
-                    bw.write_u32le(item.offs)?;
+                    bw.write_u32le(item.offs as u32)?;
                     bw.write_u32le(item.size)?;
                 }
             }
             patch_size(&mut bw, 8)?;
 
-            let mut max_frames = 0;
-            let mut max_size = 0;
+            self.patch_stream_stats()?;
+
+            let mut bw = ByteWriter::new(&mut self.fr);
+
+            // patch out useless OpenDML entries
+            bw.seek(SeekFrom::Start(self.odml_pos))?;
+            bw.write_buf(b"JUNK")?;
             for sstat in self.sstats.iter() {
-                max_frames = max_frames.max(sstat.nframes);
-                max_size = max_size.max(sstat.max_size);
-                if sstat.has_pc {
-                    bw.seek(SeekFrom::Start(sstat.strh_pos))?;
-                    bw.write_u32le(0x00010000)?;
-                }
-                bw.seek(SeekFrom::Start(sstat.strh_pos + 0x18))?;
-                bw.write_u32le(if sstat.is_video { sstat.nframes } else { 0 })?;
-                bw.write_u32le(sstat.max_size)?;
+                bw.seek(SeekFrom::Start(sstat.indx_pos))?;
+                bw.write_buf(b"JUNK")?;
             }
-            bw.seek(SeekFrom::Start(0x30))?;
-            bw.write_u32le(max_frames)?;
-            bw.seek(SeekFrom::Current(8))?;
-            bw.write_u32le(max_size)?;
 
             bw.flush()?;
 
@@ -480,5 +682,9 @@ pub fn create(name: &str) -> EncoderResult<Box<dyn OutputWriter>> {
         sstats: Vec::with_capacity(2),
         index: Vec::new(),
         data_pos: 0,
+        riff_blk: 0,
+        blk_start: 0,
+        wrote_hdr: true,
+        odml_pos: 0,
     }))
 }
